# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Static grouped assigner module."""

from openfl.component.assigner.assigner import Assigner


class StaticGroupedAssigner(Assigner):
    r"""The task assigner maintains a list of tasks.

    Also it decides the policy for which collaborator should run those tasks
    There may be many types of policies implemented, but a natural place to
    start is with a:

        - StaticGroupedAssigner :
            Given a set of task groups, and a list of
            collaborators for that group, assign tasks for of collaborators in
            the federation.
            After assigning the tasks to collaborator, those tasks should be
            carried out each round (no reassignment between rounds).
        - GroupedAssigner :
            Given task groups and a list of collaborators that
            belong to that task group, carry out tasks for each round of
            experiment.

    Attributes:
        task_groups* (list of object): Task groups to assign.

    .. note::
        \* - Plan setting.
    """

    with_selected_task_group = Assigner.with_selected_task_group

    def __init__(self, task_groups, **kwargs):
        """Initializes the StaticGroupedAssigner.

        Args:
            task_groups (list of object): Task groups to assign.
            **kwargs: Additional keyword arguments.
        """
        self.task_groups = task_groups
        super().__init__(**kwargs)

    @with_selected_task_group
    def define_task_assignments(self):
        """Define task assignments for each round and collaborator.

        This method uses the assigner function to assign tasks to
        collaborators for each round.
        It also maps tasks to their respective aggregation functions.

        Args:
            None

        Returns:
            None
        """
        cols_amount = sum([len(group["collaborators"]) for group in self.task_groups])
        authorized_cols_amount = len(self.authorized_cols)

        unique_cols = {col for group in self.task_groups for col in group["collaborators"]}
        unique_authorized_cols = set(self.authorized_cols)

        assert cols_amount == authorized_cols_amount and unique_cols == unique_authorized_cols, (
            f"Collaborators in each group must be distinct: {unique_cols}, {unique_authorized_cols}"
        )

        # Start by finding all of the tasks in all specified groups
        self.all_tasks_in_groups = list(
            {task for group in self.task_groups for task in group["tasks"]}
        )

        # Initialize the map of collaborators for a given task on a given round
        for task in self.all_tasks_in_groups:
            self.collaborators_for_task[task] = {i: [] for i in range(self.rounds)}

        for group in self.task_groups:
            group_col_list = group["collaborators"]
            self.task_group_collaborators[group["name"]] = group_col_list
            for col in group_col_list:
                # For now, we assume that collaborators have the same tasks for
                # every round
                self.collaborator_tasks[col] = dict.fromkeys(range(self.rounds), group["tasks"])
            # Now populate reverse lookup of tasks->group
            for task in group["tasks"]:
                for round_ in range(self.rounds):
                    # This should append the list of collaborators performing
                    # that task
                    self.collaborators_for_task[task][round_] += group_col_list

    def get_tasks_for_collaborator(self, collaborator_name, round_number):
        """Get tasks for a specific collaborator in a specific round.

        Args:
            collaborator_name (str): Name of the collaborator.
            round_number (int): Round number.

        Returns:
            list: List of tasks for the collaborator in the specified round.
        """
        return self.collaborator_tasks[collaborator_name][round_number]

    def get_collaborators_for_task(self, task_name, round_number):
        """Get collaborators for a specific task in a specific round.

        Args:
            task_name (str): Name of the task.
            round_number (int): Round number.

        Returns:
            list: List of collaborators for the task in the specified round.
        """
        return self.collaborators_for_task[task_name][round_number]



import sys
import os
import shutil
from logging import getLogger
from urllib.request import urlretrieve
from hashlib import sha384
from os import path, makedirs
from tqdm import tqdm
import modin.pandas as pd
import gzip
from sklearn.model_selection import train_test_split
import numpy as np

logger = getLogger(__name__)

"""HIGGS Dataset."""

URL = "https://archive.ics.uci.edu/ml/machine-learning-databases/00280/HIGGS.csv.gz"
FILENAME = "HIGGS.csv.gz"
CSV_FILENAME = "HIGGS.csv"
CSV_SHA384 = 'b8b82e11a78b81601381420878ad42ba557291f394a88dc5293e4077c8363c87429639b120e299a2a9939c1f943b6a63'
DEFAULT_PATH = path.join(os.getcwd(), 'data')

pbar = tqdm(total=None)

def report_hook(count, block_size, total_size):
    """Update progressbar."""
    if pbar.total is None and total_size:
        pbar.total = total_size
    progress_bytes = count * block_size
    pbar.update(progress_bytes - pbar.n)

def verify_sha384(file_path, expected_hash):
    """Verify the SHA-384 hash of a file."""
    sha384_hash = sha384()
    with open(file_path, 'rb') as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha384_hash.update(byte_block)
    computed_hash = sha384_hash.hexdigest()
    if computed_hash != expected_hash:
        raise ValueError(f"SHA-384 hash mismatch: expected {expected_hash}, got {computed_hash}")
    print(f"SHA-384 hash verified: {computed_hash}")

def setup_data(root: str = DEFAULT_PATH, **kwargs):
    """Initialize."""
    makedirs(root, exist_ok=True)
    filepath = path.join(root, FILENAME)
    csv_filepath = path.join(root, CSV_FILENAME)
    if not path.exists(filepath):
        urlretrieve(URL, filepath, report_hook)  # nosec
        verify_sha384(filepath, CSV_SHA384)
        # Extract the CSV file from the gzip file
        with gzip.open(filepath, 'rb') as f_in:
            with open(csv_filepath, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

def main():
    if len(sys.argv) < 2:
        raise ValueError("Provide the number of collaborators")
    src = 'higgs_data'
    if os.path.exists(src):
        shutil.rmtree(src)
    setup_data(src)
    collaborators = int(sys.argv[1])
    print("Creating splits for {} collaborators".format(collaborators))

    # Load the dataset
    higgs_data = pd.read_csv(path.join(src, CSV_FILENAME), header=None)

    # Split the dataset into features and labels
    X = higgs_data.iloc[:, 1:].values
    y = higgs_data.iloc[:, 0].values

    # Split the dataset into training and testing sets
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    # Combine X and y for train and test sets
    train_data = pd.DataFrame(data=np.column_stack((y_train, X_train)))
    test_data = pd.DataFrame(data=np.column_stack((y_test, X_test)))

    # Split the training data into parts for each collaborator
    for i in range(collaborators):
        dst = f'data/{i+1}'
        makedirs(dst, exist_ok=True)

        # Split the training data for the current collaborator
        split_train_data = train_data.iloc[i::collaborators]
        split_train_data.to_csv(path.join(dst, 'train.csv'), index=False, header=False)

        # Split the test data for the current collaborator
        split_test_data = test_data.iloc[i::collaborators]
        split_test_data.to_csv(path.join(dst, 'valid.csv'), index=False, header=False)

if __name__ == '__main__':
    main()



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Cryptography participant utilities."""

from typing import Tuple

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.x509.base import CertificateSigningRequest
from cryptography.x509.oid import NameOID


def generate_csr(
    common_name: str, server: bool = False
) -> Tuple[RSAPrivateKey, CertificateSigningRequest]:
    """Issue a Certificate Signing Request (CSR) for a server or a client.

    Args:
        common_name (str): The common name for the certificate.
        server (bool, optional): A flag to indicate if the CSR is for a server.
            If False, the CSR is for a client. Defaults to False.

    Returns:
        Tuple[RSAPrivateKey, CertificateSigningRequest]: The private key and
            the CSR.
    """
    # Generate private key
    private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=3072, backend=default_backend()
    )

    builder = x509.CertificateSigningRequestBuilder()
    subject = x509.Name(
        [
            x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        ]
    )
    builder = builder.subject_name(subject)
    builder = builder.add_extension(
        x509.BasicConstraints(ca=False, path_length=None),
        critical=True,
    )
    if server:
        builder = builder.add_extension(
            x509.ExtendedKeyUsage([x509.ExtendedKeyUsageOID.SERVER_AUTH]),
            critical=True,
        )

    else:
        builder = builder.add_extension(
            x509.ExtendedKeyUsage([x509.ExtendedKeyUsageOID.CLIENT_AUTH]),
            critical=True,
        )

    builder = builder.add_extension(
        x509.KeyUsage(
            digital_signature=True,
            key_encipherment=True,
            data_encipherment=False,
            key_agreement=False,
            content_commitment=False,
            key_cert_sign=False,
            crl_sign=False,
            encipher_only=False,
            decipher_only=False,
        ),
        critical=True,
    )

    builder = builder.add_extension(
        x509.SubjectAlternativeName([x509.DNSName(common_name)]), critical=False
    )

    # Sign the CSR
    csr = builder.sign(
        private_key=private_key,
        algorithm=hashes.SHA384(),
        backend=default_backend(),
    )

    return private_key, csr



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.component.collaborator package."""

# FIXME: Too much recursion.
from openfl.experimental.workflow.component.collaborator.collaborator import Collaborator



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Federated averaging with secure aggregation module."""

import numpy as np

from openfl.interface.aggregation_functions.weighted_average import WeightedAverage
from openfl.utilities import LocalTensor


class SecureWeightedAverage(WeightedAverage):
    """Weighted average with secure aggregation."""

    def __init__(self):
        super().__init__()
        self._private_masks, self._shared_masks = None, None
        self.private_seeds, self.agreed_keys, self.col_indices = None, None, None

    def call(self, local_tensors, db_iterator, *_) -> np.ndarray:
        """Aggregate tensors.

        Args:
            local_tensors (list[openfl.utilities.LocalTensor]): List of local
                tensors to aggregate.
            db_iterator: iterator over history of all tensors. Columns:
                - 'tensor_name': name of the tensor.
                    Examples for `torch.nn.Module`s: 'conv1.weight','fc2.bias'.
                - 'round': 0-based number of round corresponding to this
                    tensor.
                - 'tags': tuple of tensor tags. Tags that can appear:
                    - 'model' indicates that the tensor is a model parameter.
                    - 'trained' indicates that tensor is a part of a training
                        result.
                        These tensors are passed to the aggregator node after
                        local learning.
                    - 'aggregated' indicates that tensor is a result of
                        aggregation.
                        These tensors are sent to collaborators for the next
                        round.
                    - 'delta' indicates that value is a difference between
                        rounds for a specific tensor.
                    also one of the tags is a collaborator name
                    if it corresponds to a result of a local task.

                - 'nparray': value of the tensor.
            tensor_name: name of the tensor
            fl_round: round number
            tags: tuple of tags for this tensor
        Returns:
            np.ndarray: aggregated tensor
        """
        # Generate masks for the collaborators if not already done.
        self._generate_masks(db_iterator)
        # Calaculate the weighted avreage of collaborator masks.
        weighted_mask = self._calculcate_weighted_mask_average(self._private_masks, local_tensors)
        # Get weighted average for shared tensors.
        tensor_avg = super().call(local_tensors)
        # Subtract weighted average of masks from the tensor average.
        return np.subtract(np.subtract(tensor_avg, weighted_mask), self._shared_masks)

    def _generate_masks(self, db_iterator):
        """
        Generate shared and private masks for secure aggregation.

        This method processes a database iterator to extract private seeds,
        agreed keys, and column indices, which are then used to generate
        shared and private masks.

        Args:
            db_iterator (iterator): An iterator over the database items
                containing tensors with tags, tensor names, and numpy arrays.

        Raises:
            KeyError: If the required keys are not found in the database items.

        Notes:
            - The shared masks are calculated using the agreed keys.
            - The private masks are generated for each collaborator using
                their private seeds.
            - The private masks are stored in a dictionary with the
                collaborator's name as the key.
        """
        from openfl.utilities.secagg import (
            calculate_shared_mask,
            pseudo_random_generator,
        )

        if self._shared_masks and self._private_masks:
            return

        # Get all required values from tensor db.
        self._get_secagg_items_from_db(db_iterator)
        if not self._shared_masks:
            # Calculate shared mask
            self._shared_masks = calculate_shared_mask(self.agreed_keys)

        if not self._private_masks:
            # Create a dict with collaborator index and their name.
            # This dict is used to map private masks to the collaborator name
            # as they are stored with collaborator index in the db.
            col_idx = {}
            for col in self.col_indices:
                col_idx[col[1]] = col[0]

            # Generate private masks for each collaborator.
            self._private_masks = {}
            for seed in self.private_seeds:
                # col_name: col_private_mask
                self._private_masks[col_idx[seed[0]]] = pseudo_random_generator(seed[1])

            del col_idx

    def _calculcate_weighted_mask_average(
        self, private_masks: dict, local_tensors: list[LocalTensor]
    ):
        """
        Calculate the weighted mask average for the given local tensors using
        their private masks and weight for their respective tensors.

        Args:
            private_masks (dict): A dictionary where keys are collaborator
                names and values are tuples, with the second element being the
                mask for that colaborator.
            local_tensors (list): A list of tensors, where each tensor has
                attributes 'col_name' and 'weight'.

        Returns:
            numpy.ndarray: The average mask calculated as the weighted
                average of the masks.
        """
        weights = []
        masks = []
        # Create a list of private masks and weights for the collaborators
        # whose tensors are being aggregated.
        for tensor in local_tensors:
            col_name = tensor.col_name
            weights.append(tensor.weight)
            masks.append(private_masks[col_name])

        # Calculate weighted mask using the masks and weights where each index
        # in the lists represents a single collaborator.
        weighted_mask = np.average(masks, weights=weights, axis=0)

        del weights
        del masks

        return weighted_mask

    def _get_secagg_items_from_db(self, db_iterator):
        """
        Extracts secure aggregation items from a database iterator.
        It retrieves the private seeds, agreed keys, and column indices from
        the database items.

        Args:
            db_iterator (iterable): An iterator that yields database items.
                Each item is expected to be a dictionary with keys "tags",
                "tensor_name", and "nparray".

        Raises:
            KeyError: If any of the required keys ("tags", "tensor_name",
                "nparray") are missing in an item.
        """
        for item in db_iterator:
            if "tags" in item and item["tags"] == ("secagg",):
                if item["tensor_name"] == "private_seeds":
                    self.private_seeds = item["nparray"]
                elif item["tensor_name"] == "agreed_keys":
                    self.agreed_keys = item["nparray"]
                elif item["tensor_name"] == "indices":
                    self.col_indices = item["nparray"]



# Copyright (C) 2024 Intel Corporation
# Licensed subject to the terms of the separately executed evaluation license agreement between
# Intel Corporation and you.

"""You may copy this file as the starting point of your own model."""
import numpy as np
import xgboost as xgb

from openfl.federated import XGBoostTaskRunner
from openfl.utilities import Metric
from sklearn.metrics import accuracy_score


class XGBoostRunner(XGBoostTaskRunner):
    """
    A class to run XGBoost training and validation tasks.

    This class inherits from XGBoostTaskRunner and provides methods to train and validate
    an XGBoost model using federated learning.

    Attributes:
        bst (xgb.Booster): The XGBoost model.
        params (dict): Parameters for the XGBoost model.
        num_rounds (int): Number of boosting rounds.
    """
    def __init__(self, params=None, num_rounds=1, **kwargs):
        """
        Initialize the XGBoostRunner.

        Args:
            params (dict, optional): Parameters for the XGBoost model. Defaults to None.
            num_rounds (int, optional): Number of boosting rounds. Defaults to 1.
            **kwargs: Additional arguments to pass to the function.
        """
        super().__init__(**kwargs)

        self.bst = None
        self.params = params
        self.num_rounds = num_rounds

    def train_(self, data) -> Metric:
        """
        Train the XGBoost model.

        Args:
            train_dataloader (dict): A dictionary containing the training data with keys 'dmatrix'.

        Returns:
            Metric: A Metric object containing the training loss.
        """
        dtrain = data['dmatrix']
        evals = [(dtrain, 'train')]
        evals_result = {}

        self.bst = xgb.train(self.params, dtrain, self.num_rounds, xgb_model=self.bst,
                             evals=evals, evals_result=evals_result, verbose_eval=False)

        loss = evals_result['train']['logloss'][-1]
        return Metric(name=self.params['eval_metric'], value=np.array(loss))

    def validate_(self, data) -> Metric:
        """
        Validate the XGBoost model.

        Args:
            validation_dataloader (dict): A dictionary containing the validation data with keys 'dmatrix' and 'labels'.

        Returns:
            Metric: A Metric object containing the validation accuracy.
        """
        dtest = data['dmatrix']
        y_test = data['labels']
        preds = self.bst.predict(dtest)
        y_pred_binary = np.where(preds > 0.5, 1, 0)
        acc = accuracy_score(y_test, y_pred_binary)

        return Metric(name="accuracy", value=np.array(acc))



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""GaNDLFTaskRunner module."""

import logging
import os
from copy import deepcopy
from typing import Union

import numpy as np
import torch as pt
import yaml
from GANDLF.compute.forward_pass import validate_network
from GANDLF.compute.generic import create_pytorch_objects
from GANDLF.compute.training_loop import train_network
from GANDLF.config_manager import ConfigManager

from openfl.federated.task.runner import TaskRunner
from openfl.utilities import TensorKey
from openfl.utilities.split import split_tensor_dict_for_holdouts

logger = logging.getLogger(__name__)


class GaNDLFTaskRunner(TaskRunner):
    """GaNDLF Model class for Federated Learning.

    This class provides methods to manage and manipulate GaNDLF models in a
    federated learning context.

    Attributes:
        build_model (function or class): Function or Class to build the model.
        lambda_opt (function): Lambda function for the optimizer.
        model (Model): The built model.
        optimizer (Optimizer): Optimizer for the model.
        scheduler (Scheduler): Scheduler for the model.
        params (Parameters): Parameters for the model.
        device (str): Device for the model.
        training_round_completed (bool): Whether the training round has been
            completed.
        required_tensorkeys_for_function (dict): Required tensorkeys for
            function.
        tensor_dict_split_fn_kwargs (dict): Keyword arguments for the tensor
            dict split function.
    """

    def __init__(
        self,
        gandlf_config: Union[str, dict] = None,
        device: str = None,
        **kwargs,
    ):
        """Initializes the GaNDLFTaskRunner object.

        Sets up the initial state of the GaNDLFTaskRunner object, initializing
        various components needed for the federated model.
        Args:
            gandlf_config (Union[str, dict], optional): GaNDLF configuration.
                Can be a string (file path) or a dictionary. Defaults to None.
            device (str, optional): Compute device. Defaults to None
                (default="cpu").
            **kwargs: Additional parameters to pass to the function.
        """
        super().__init__(**kwargs)

        assert bool(gandlf_config), "gandlf_config must be specified"

        # allow pass-through of a gandlf config as a file or a dict

        train_csv = self.data_loader.train_csv
        val_csv = self.data_loader.val_csv

        if isinstance(gandlf_config, str) and os.path.exists(gandlf_config):
            gandlf_config = yaml.safe_load(open(gandlf_config, "r"))

        gandlf_config = ConfigManager(gandlf_config)

        (
            model,
            optimizer,
            train_loader,
            val_loader,
            scheduler,
            params,
        ) = create_pytorch_objects(
            gandlf_config, train_csv=train_csv, val_csv=val_csv, device=device
        )
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.params = params
        self.device = device

        # pass the actual dataloaders to the wrapper loader
        self.data_loader.set_dataloaders(train_loader, val_loader)

        self.training_round_completed = False

        self.required_tensorkeys_for_function = {}

        # FIXME: why isn't this initial call in runner_pt?
        self.initialize_tensorkeys_for_functions(with_opt_vars=False)

        # overwrite attribute to account for one optimizer param (in every
        # child model that does not overwrite get and set tensordict) that is
        # not a numpy array
        self.tensor_dict_split_fn_kwargs.update({"holdout_tensor_names": ["__opt_state_needed"]})

    def rebuild_model(self, round_num, input_tensor_dict, validation=False):
        """Parse tensor names and update weights of model. Handles the
        optimizer treatment.

        Args:
            round_num: The current round number.
            input_tensor_dict (dict): The input tensor dictionary used to
                update the weights of the model.
            validation (bool, optional): A flag indicating whether the model
                is in validation. Defaults to False.

        Returns:
            None
        """

        if self.opt_treatment == "RESET":
            self.reset_opt_vars()
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)
        elif (
            self.training_round_completed
            and self.opt_treatment == "CONTINUE_GLOBAL"
            and not validation
        ):
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=True)
        else:
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)

    def validate(self, col_name, round_num, input_tensor_dict, use_tqdm=False, **kwargs):
        """Validate.
        Run validation of the model on the local data.
        Args:
            col_name (str): Name of the collaborator.
            round_num (int): Current round number.
            input_tensor_dict (dict): Required input tensors (for model).
            use_tqdm (bool, optional): Use tqdm to print a progress bar.
                Defaults to False.
            **kwargs: Key word arguments passed to GaNDLF main_run.

        Returns:
            output_tensor_dict (dict): Tensors to send back to the aggregator.
            {} (dict): Tensors to maintain in the local TensorDB.
        """
        self.rebuild_model(round_num, input_tensor_dict, validation=True)
        self.model.eval()

        epoch_valid_loss, epoch_valid_metric = validate_network(
            self.model,
            self.data_loader.val_dataloader,
            self.scheduler,
            self.params,
            round_num,
            mode="validation",
        )

        logger.info(epoch_valid_loss)
        logger.info(epoch_valid_metric)

        origin = col_name
        suffix = "validate"
        if kwargs["apply"] == "local":
            suffix += "_local"
        else:
            suffix += "_agg"
        tags = ("metric", suffix)

        output_tensor_dict = {}
        valid_loss_tensor_key = TensorKey("valid_loss", origin, round_num, True, tags)
        output_tensor_dict[valid_loss_tensor_key] = np.array(epoch_valid_loss)
        for k, v in epoch_valid_metric.items():
            tensor_key = TensorKey(f"valid_{k}", origin, round_num, True, tags)
            output_tensor_dict[tensor_key] = np.array(v)

        # Empty list represents metrics that should only be stored locally
        return output_tensor_dict, {}

    def train(
        self,
        col_name,
        round_num,
        input_tensor_dict,
        use_tqdm=False,
        epochs=1,
        **kwargs,
    ):
        """Train batches.
        Train the model on the requested number of batches.
        Args:
            col_name (str): Name of the collaborator.
            round_num (int): Current round number.
            input_tensor_dict (dict): Required input tensors (for model).
            use_tqdm (bool, optional): Use tqdm to print a progress bar.
                Defaults to False.
            epochs (int, optional): The number of epochs to train. Defaults to 1.
            **kwargs: Key word arguments passed to GaNDLF main_run.

        Returns:
            global_tensor_dict (dict): Tensors to send back to the aggregator.
            local_tensor_dict (dict): Tensors to maintain in the local
                TensorDB.
        """
        self.rebuild_model(round_num, input_tensor_dict)
        # set to "training" mode
        self.model.train()
        for epoch in range(epochs):
            logger.info("Run %s epoch of %s round", epoch, round_num)
            # FIXME: do we want to capture these in an array
            # rather than simply taking the last value?
            epoch_train_loss, epoch_train_metric = train_network(
                self.model,
                self.data_loader.train_dataloader,
                self.optimizer,
                self.params,
            )

        # output model tensors (Doesn't include TensorKey)
        tensor_dict = self.get_tensor_dict(with_opt_vars=(self.opt_treatment == "CONTINUE_GLOBAL"))

        metric_dict = {"loss": epoch_train_loss}
        for k, v in epoch_train_metric.items():
            metric_dict[f"train_{k}"] = v

        # Return global_tensor_dict, local_tensor_dict
        # is this even pt-specific really?
        global_tensor_dict, local_tensor_dict = create_tensorkey_dicts(
            tensor_dict,
            metric_dict,
            col_name,
            round_num,
            self.tensor_dict_split_fn_kwargs,
        )

        # Update the required tensors if they need to be pulled from the
        # aggregator
        # TODO this logic can break if different collaborators have different
        # roles between rounds.
        # For example, if a collaborator only performs validation in the first
        # round but training in the second, it has no way of knowing the
        # optimizer state tensor names to request from the aggregator because
        # these are only created after training occurs. A work around could
        # involve doing a single epoch of training on random data to get the
        # optimizer names, and then throwing away the model.
        if self.opt_treatment == "CONTINUE_GLOBAL":
            self.initialize_tensorkeys_for_functions(with_opt_vars=True)

        # This will signal that the optimizer values are now present,
        # and can be loaded when the model is rebuilt
        self.training_round_completed = True

        # Return global_tensor_dict, local_tensor_dict
        return global_tensor_dict, local_tensor_dict

    def get_tensor_dict(self, with_opt_vars=False):
        """Return the tensor dictionary.

        Args:
            with_opt_vars (bool): Return the tensor dictionary including the
                optimizer tensors (Default=False).

        Returns:
            state (dict): Tensor dictionary {**dict, **optimizer_dict}
        """
        # Gets information regarding tensor model layers and optimizer state.
        # FIXME: self.parameters() instead? Unclear if load_state_dict() or
        # simple assignment is better
        # for now, state dict gives us names which is good
        # FIXME: do both and sanity check each time?

        state = to_cpu_numpy(self.model.state_dict())

        if with_opt_vars:
            opt_state = _get_optimizer_state(self.optimizer)
            state = {**state, **opt_state}

        return state

    def _get_weights_names(self, with_opt_vars=False):
        """Get the names of the weights.

        Args:
            with_opt_vars (bool, optional): Include the optimizer variables.
                Defaults to False.

        Returns:
            list: List of weight names.
        """
        # Gets information regarding tensor model layers and optimizer state.
        # FIXME: self.parameters() instead? Unclear if load_state_dict() or
        # simple assignment is better
        # for now, state dict gives us names which is good
        # FIXME: do both and sanity check each time?

        state = self.model.state_dict().keys()

        if with_opt_vars:
            opt_state = _get_optimizer_state(self.model.optimizer)
            state += opt_state.keys()

        return state

    def set_tensor_dict(self, tensor_dict, with_opt_vars=False):
        """Set the tensor dictionary.

        Args:
            tensor_dict (dict): The tensor dictionary.
            with_opt_vars (bool, optional): Include the optimizer tensors.
                Defaults to False.
        """
        set_pt_model_from_tensor_dict(self.model, tensor_dict, self.device, with_opt_vars)

    def get_optimizer(self):
        """Get the optimizer of this instance.

        Returns:
            Optimizer: The optimizer of this instance.
        """
        return self.optimizer

    def get_required_tensorkeys_for_function(self, func_name, **kwargs):
        """Get the required tensors for specified function that could be called
        as part of a task.

        By default, this is just all of the layers and optimizer of the model.

        Args:
            func_name (str): Function name.
            **kwargs: Additional keyword arguments.

        Returns:
            required_tensorkeys_for_function (list): List of required
                TensorKey.
        """
        if func_name == "validate":
            local_model = "apply=" + str(kwargs["apply"])
            return self.required_tensorkeys_for_function[func_name][local_model]
        else:
            return self.required_tensorkeys_for_function[func_name]

    def initialize_tensorkeys_for_functions(self, with_opt_vars=False):
        """Set the required tensors for all publicly accessible task methods.

        By default, this is just all of the layers and optimizer of the model.
        Custom tensors should be added to this function.

        Args:
            with_opt_vars (bool, optional): Include the optimizer tensors.
                Defaults to False.
        """
        # TODO there should be a way to programmatically iterate through
        #  all of the methods in the class and declare the tensors.
        # For now this is done manually

        output_model_dict = self.get_tensor_dict(with_opt_vars=with_opt_vars)
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            output_model_dict, **self.tensor_dict_split_fn_kwargs
        )
        if not with_opt_vars:
            global_model_dict_val = global_model_dict
            local_model_dict_val = local_model_dict
        else:
            output_model_dict = self.get_tensor_dict(with_opt_vars=False)
            global_model_dict_val, local_model_dict_val = split_tensor_dict_for_holdouts(
                output_model_dict,
                **self.tensor_dict_split_fn_kwargs,
            )

        self.required_tensorkeys_for_function["train"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict
        ]
        self.required_tensorkeys_for_function["train"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict
        ]

        # Validation may be performed on local or aggregated (global) model,
        # so there is an extra lookup dimension for kwargs
        self.required_tensorkeys_for_function["validate"] = {}
        # TODO This is not stateless. The optimizer will not be
        self.required_tensorkeys_for_function["validate"]["apply=local"] = [
            TensorKey(tensor_name, "LOCAL", 0, False, ("trained",))
            for tensor_name in {**global_model_dict_val, **local_model_dict_val}
        ]
        self.required_tensorkeys_for_function["validate"]["apply=global"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict_val
        ]
        self.required_tensorkeys_for_function["validate"]["apply=global"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict_val
        ]

    def load_native(
        self,
        filepath,
        model_state_dict_key="model_state_dict",
        optimizer_state_dict_key="optimizer_state_dict",
        **kwargs,
    ):
        """
        Load model and optimizer states from a pickled file specified by \
        filepath. model_/optimizer_state_dict args can be specified if needed. \
        Uses pt.load().

        Args:
            filepath (str): Path to pickle file created by pt.save().
            model_state_dict_key (str, optional): Key for model state dict in
                pickled file. Defaults to 'model_state_dict'.
            optimizer_state_dict_key (str, optional): Key for optimizer state
                dict in picked file. Defaults to 'optimizer_state_dict'.
            **kwargs: Additional keyword arguments.
        """
        pickle_dict = pt.load(filepath, weights_only=True)  # nosec B614

        if model_state_dict_key in pickle_dict and optimizer_state_dict_key in pickle_dict:
            self.load_state_dict(pickle_dict[model_state_dict_key])
            self.optimizer.load_state_dict(pickle_dict[optimizer_state_dict_key])
        else:
            self.load_state_dict(pickle_dict)

    def save_native(
        self,
        filepath,
        model_state_dict_key="model_state_dict",
        optimizer_state_dict_key="optimizer_state_dict",
        **kwargs,
    ):
        """
        Save model and optimizer states in a picked file specified by the \
        filepath. model_/optimizer_state_dicts are stored in the keys provided. \
        Uses pt.save().

        Args:
            filepath (str): Path to pickle file to be created by pt.save().
                By default, model will be saved as `*.pt`
            model_state_dict_key (str, optional): Key for model state dict in
                pickled file. Defaults to 'model_state_dict'.
            optimizer_state_dict_key (str, optional): Key for optimizer state
                dict in picked file. Defaults to 'optimizer_state_dict'.
            **kwargs: Additional keyword arguments.
        """
        pickle_dict = {
            model_state_dict_key: self.model.state_dict(),
            optimizer_state_dict_key: self.optimizer.state_dict(),
        }
        if "." not in str(filepath).split(os.sep)[-1]:
            filepath = str(filepath) + ".pt"

        pt.save(pickle_dict, filepath)
        return filepath

    def reset_opt_vars(self):
        """Reset optimizer variables."""
        pass


def create_tensorkey_dicts(
    tensor_dict,
    metric_dict,
    col_name,
    round_num,
    tensor_dict_split_fn_kwargs,
):
    """Create dictionaries of TensorKeys for global and local tensors.

    Args:
        tensor_dict (dict): Dictionary of tensors.
        metric_dict (dict): Dictionary of metrics.
        col_name (str): Name of the collaborator.
        round_num (int): Current round number.
        tensor_dict_split_fn_kwargs (dict): Keyword arguments for the tensor
            dict split function.

    Returns:
        global_tensor_dict (dict): Dictionary of global TensorKeys.
        local_tensor_dict (dict): Dictionary of local TensorKeys.
    """
    origin = col_name
    tags = ("trained",)
    output_metric_dict = {}
    for k, v in metric_dict.items():
        tk = TensorKey(k, origin, round_num, True, ("metric",))
        output_metric_dict[tk] = np.array(v)

    global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
        tensor_dict, **tensor_dict_split_fn_kwargs
    )

    # Create global tensorkeys
    global_tensorkey_model_dict = {
        TensorKey(tensor_name, origin, round_num, False, tags): nparray
        for tensor_name, nparray in global_model_dict.items()
    }
    # Create tensorkeys that should stay local
    local_tensorkey_model_dict = {
        TensorKey(tensor_name, origin, round_num, False, tags): nparray
        for tensor_name, nparray in local_model_dict.items()
    }
    # The train/validate aggregated function of the next round will look
    # for the updated model parameters.
    # This ensures they will be resolved locally
    next_local_tensorkey_model_dict = {
        TensorKey(tensor_name, origin, round_num + 1, False, ("model",)): nparray
        for tensor_name, nparray in local_model_dict.items()
    }

    global_tensor_dict = {**output_metric_dict, **global_tensorkey_model_dict}
    local_tensor_dict = {
        **local_tensorkey_model_dict,
        **next_local_tensorkey_model_dict,
    }

    return global_tensor_dict, local_tensor_dict


def set_pt_model_from_tensor_dict(model, tensor_dict, device, with_opt_vars=False):
    """Set the tensor dictionary for the PyTorch model.

    Args:
        model (Model): The PyTorch model.
        tensor_dict (dict): Tensor dictionary.
        device (str): Device for the model.
        with_opt_vars (bool, optional): Include the optimizer tensors.
            Defaults to False.
    """
    # Sets tensors for model layers and optimizer state.
    # FIXME: model.parameters() instead? Unclear if load_state_dict() or
    #  simple assignment is better
    # for now, state dict gives us names, which is good
    # FIXME: do both and sanity check each time?

    new_state = {}
    # Grabbing keys from model's state_dict helps to confirm we have
    # everything
    for k in model.state_dict():
        new_state[k] = pt.from_numpy(tensor_dict.pop(k)).to(device)

    # set model state
    model.load_state_dict(new_state)

    if with_opt_vars:
        # see if there is state to restore first
        if tensor_dict.pop("__opt_state_needed") == "true":
            _set_optimizer_state(model.get_optimizer(), device, tensor_dict)

        # sanity check that we did not record any state that was not used
        assert len(tensor_dict) == 0


def _derive_opt_state_dict(opt_state_dict):
    """Separate optimizer tensors from the tensor dictionary.

    Flattens the optimizer state dict so as to have key, value pairs with
    values as numpy arrays.
    The keys have sufficient info to restore opt_state_dict using
    expand_derived_opt_state_dict.

    Args:
        opt_state_dict (dict): Optimizer state dictionary.

    Returns:
        derived_opt_state_dict (dict): Optimizer state dictionary.
    """
    derived_opt_state_dict = {}

    # Determine if state is needed for this optimizer.
    if len(opt_state_dict["state"]) == 0:
        derived_opt_state_dict["__opt_state_needed"] = "false"
        return derived_opt_state_dict

    derived_opt_state_dict["__opt_state_needed"] = "true"

    # Using one example state key, we collect keys for the corresponding
    # dictionary value.
    example_state_key = opt_state_dict["param_groups"][0]["params"][0]
    example_state_subkeys = set(opt_state_dict["state"][example_state_key].keys())

    # We assume that the state collected for all params in all param groups is
    # the same.
    # We also assume that whether or not the associated values to these state
    # subkeys is a tensor depends only on the subkey.
    # Using assert statements to break the routine if these assumptions are
    # incorrect.
    for state_key in opt_state_dict["state"].keys():
        assert example_state_subkeys == set(opt_state_dict["state"][state_key].keys())
        for state_subkey in example_state_subkeys:
            assert isinstance(
                opt_state_dict["state"][example_state_key][state_subkey],
                pt.Tensor,
            ) == isinstance(opt_state_dict["state"][state_key][state_subkey], pt.Tensor)

    state_subkeys = list(opt_state_dict["state"][example_state_key].keys())

    # Tags will record whether the value associated to the subkey is a
    # tensor or not.
    state_subkey_tags = []
    for state_subkey in state_subkeys:
        if isinstance(opt_state_dict["state"][example_state_key][state_subkey], pt.Tensor):
            state_subkey_tags.append("istensor")
        else:
            state_subkey_tags.append("")
    state_subkeys_and_tags = list(zip(state_subkeys, state_subkey_tags))

    # Forming the flattened dict, using a concatenation of group index,
    # subindex, tag, and subkey inserted into the flattened dict key -
    # needed for reconstruction.
    nb_params_per_group = []
    for group_idx, group in enumerate(opt_state_dict["param_groups"]):
        for idx, param_id in enumerate(group["params"]):
            for subkey, tag in state_subkeys_and_tags:
                if tag == "istensor":
                    new_v = opt_state_dict["state"][param_id][subkey].cpu().numpy()
                else:
                    new_v = np.array([opt_state_dict["state"][param_id][subkey]])
                derived_opt_state_dict[f"__opt_state_{group_idx}_{idx}_{tag}_{subkey}"] = new_v
        nb_params_per_group.append(idx + 1)
    # group lengths are also helpful for reconstructing
    # original opt_state_dict structure
    derived_opt_state_dict["__opt_group_lengths"] = np.array(nb_params_per_group)

    return derived_opt_state_dict


def expand_derived_opt_state_dict(derived_opt_state_dict, device):
    """Expand the optimizer state dictionary.

    Takes a derived opt_state_dict and creates an opt_state_dict suitable as
    input for load_state_dict for restoring optimizer state.
    Reconstructing state_subkeys_and_tags using the example key prefix,
    "__opt_state_0_0_", certain to be present.

    Args:
        derived_opt_state_dict (dict): Derived optimizer state dictionary.
        device (str): Device for the model.

    Returns:
        opt_state_dict (dict): Expanded optimizer state dictionary.
    """
    state_subkeys_and_tags = []
    for key in derived_opt_state_dict:
        if key.startswith("__opt_state_0_0_"):
            stripped_key = key[16:]
            if stripped_key.startswith("istensor_"):
                this_tag = "istensor"
                subkey = stripped_key[9:]
            else:
                this_tag = ""
                subkey = stripped_key[1:]
            state_subkeys_and_tags.append((subkey, this_tag))

    opt_state_dict = {"param_groups": [], "state": {}}
    nb_params_per_group = list(derived_opt_state_dict.pop("__opt_group_lengths").astype(np.int32))

    # Construct the expanded dict.
    for group_idx, nb_params in enumerate(nb_params_per_group):
        these_group_ids = [f"{group_idx}_{idx}" for idx in range(nb_params)]
        opt_state_dict["param_groups"].append({"params": these_group_ids})
        for this_id in these_group_ids:
            opt_state_dict["state"][this_id] = {}
            for subkey, tag in state_subkeys_and_tags:
                flat_key = f"__opt_state_{this_id}_{tag}_{subkey}"
                if tag == "istensor":
                    new_v = pt.from_numpy(derived_opt_state_dict.pop(flat_key))
                else:
                    # Here (for currently supported optimizers) the subkey
                    # should be 'step' and the length of array should be one.
                    assert subkey == "step"
                    assert len(derived_opt_state_dict[flat_key]) == 1
                    new_v = int(derived_opt_state_dict.pop(flat_key))
                opt_state_dict["state"][this_id][subkey] = new_v

    # sanity check that we did not miss any optimizer state
    assert len(derived_opt_state_dict) == 0

    return opt_state_dict


def _get_optimizer_state(optimizer):
    """Get the state of the optimizer.

    Args:
        optimizer (Optimizer): Optimizer.

    Returns:
        derived_opt_state_dict (dict): State of the optimizer.
    """
    opt_state_dict = deepcopy(optimizer.state_dict())

    # Optimizer state might not have some parts representing frozen parameters
    # So we do not synchronize them
    param_keys_with_state = set(opt_state_dict["state"].keys())
    for group in opt_state_dict["param_groups"]:
        local_param_set = set(group["params"])
        params_to_sync = local_param_set & param_keys_with_state
        group["params"] = sorted(params_to_sync)

    derived_opt_state_dict = _derive_opt_state_dict(opt_state_dict)

    return derived_opt_state_dict


def _set_optimizer_state(optimizer, device, derived_opt_state_dict):
    """Set the state of the optimizer.

    Args:
        optimizer (Optimizer): Optimizer.
        device (str): Device for the model.
        derived_opt_state_dict (dict): Derived optimizer state dictionary.
    """
    temp_state_dict = expand_derived_opt_state_dict(derived_opt_state_dict, device)

    # FIXME: Figure out whether or not this breaks learning rate
    #  scheduling and the like.
    # Setting default values.
    # All optimizer.defaults are considered as not changing over course of
    # training.
    for group in temp_state_dict["param_groups"]:
        for k, v in optimizer.defaults.items():
            group[k] = v

    optimizer.load_state_dict(temp_state_dict)


def to_cpu_numpy(state):
    """Convert state to CPU as Numpy array.

    Args:
        state (State): State to be converted.

    Returns:
        state (dict): State as Numpy array.
    """
    # deep copy so as to decouple from active model
    state = deepcopy(state)

    for k, v in state.items():
        # When restoring, we currently assume all values are tensors.
        if not pt.is_tensor(v):
            raise ValueError(
                "We do not currently support non-tensors coming from model.state_dict()"
            )
        # get as a numpy array, making sure is on cpu
        state[k] = v.cpu().numpy()
    return state



# Copyright (C) 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from collections.abc import Iterable
from logging import getLogger
import os
import sys


from openfl.federated import PyTorchDataLoader
import numpy as np
from openfl.federated.data.sources.torch.verifiable_map_style_image_folder import VerifiableImageFolder
from openfl.federated.data.sources.data_sources_json_parser import DataSourcesJsonParser
from openfl.utilities.path_check import is_directory_traversal
import torch
from torch.utils.data import random_split
from torchvision.transforms import ToTensor


logger = getLogger(__name__)


class PyTorchHistologyVerifiableDataLoader(PyTorchDataLoader):
    """PyTorch data loader for Histology dataset."""

    def __init__(self, data_path, batch_size, **kwargs):
        """Instantiate the data object.

        Args:
            data_path: The file path to the data
            batch_size: The batch size of the data loader
            **kwargs: Additional arguments, passed to super init
             and load_mnist_shard
        """
        super().__init__(batch_size, random_seed=0, **kwargs)

        # Set Histology-specific default attributes
        self.feature_shape = [3, 150, 150]
        self.num_classes = 8

        # If data_path is None, this is being used for model initialization only
        if data_path is None:
            logger.info("Initializing dataloader for model creation only (no data loading)")
            return

        verifible_dataset_info = self.get_verifiable_dataset_info(data_path)

        hash_file_path = os.path.join(data_path, "hash.txt")
        verify_dataset_items = False
        if os.path.isfile(hash_file_path):
            logger.info(f"Found hash file at: {hash_file_path}. Verifying dataset...")
            verify_dataset_items = True
            with open(hash_file_path, "r", encoding="utf-8") as file:
                hash_value = file.read()
            dataset_valid = verifible_dataset_info.verify_dataset(root_hash=hash_value)
            if not dataset_valid:
                logger.error("The dataset is not valid.")
                sys.exit(1)
            else:
                logger.info("The dataset is valid.")

        _, num_classes, X_train, y_train, X_valid, y_valid = load_histology_shard(
            verifible_dataset_info=verifible_dataset_info, verify_dataset_items=verify_dataset_items, **kwargs)

        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid

        self.num_classes = num_classes


    def get_feature_shape(self):
        """Returns the shape of an example feature array.

        Returns:
            list: The shape of an example feature array [3, 150, 150] for Histology images.
        """
        return self.feature_shape

    def get_num_classes(self):
        """Returns the number of classes for classification tasks.

        Returns:
            int: The number of classes (8 for Histology dataset).
        """
        return self.num_classes

    def get_verifiable_dataset_info(self, data_path):
        """
        Parse dataset information from `datasources.json` in the given directory.

        Args:
            data_path (str): Path to the directory containing `datasources.json`.

        Returns:
            VerifiableDatasetInfo: Parsed dataset information.

        Raises:
            SystemExit: If `data_path` is invalid or missing `datasources.json`.
        """
        """Return the verifiable dataset info object for the given data sources."""
        if data_path and is_directory_traversal(data_path):
            logger.error("Data path is out of the openfl workspace scope.")
        if not os.path.isdir(data_path):
            logger.error("The data path must be a directory.")
            sys.exit(1)

        datasources_json_path = os.path.join(data_path, "datasources.json")
        if not os.path.isfile(datasources_json_path):
            logger.error("The directory must contain a file named 'datasources.json' at the first level.")
            sys.exit(1)
        with open(datasources_json_path, "r", encoding="utf-8") as file:
            data = file.read()
        return DataSourcesJsonParser.parse(data)


class HistologyDataset(VerifiableImageFolder):
    """Colorectal Histology Dataset."""

    def __init__(self, **kwargs) -> None:
        """Initialize."""
        super().__init__(**kwargs)

    def __getitem__(self, index):
        """Allow getting items by slice index."""
        if isinstance(index, Iterable):
            return [super().__getitem__(i) for i in index]
        else:
            return super().__getitem__(index)


def _load_raw_data(verifiable_dataset_info, verify_dataset_items=False, train_split_ratio=0.8, **kwargs):
    """
    Load the raw data by shard.

    Returns tuples of the dataset shard divided into training and validation.

    Args:
        data_path (str): The path to the dataset.

    Returns:
        2 tuples: (image, label) of the training, validation dataset
    """
    dataset = HistologyDataset(
        vds=verifiable_dataset_info,
        transform=ToTensor(),
        verify_dataset_items=verify_dataset_items
    )
    n_train = int(train_split_ratio * len(dataset))
    n_valid = len(dataset) - n_train
    ds_train, ds_val = random_split(
        dataset, lengths=[n_train, n_valid], generator=torch.manual_seed(0))

    # create the shards
    X_train, y_train = list(zip(*ds_train))
    X_train, y_train = np.stack(X_train), np.array(y_train)

    X_valid, y_valid = list(zip(*ds_val))
    X_valid, y_valid = np.stack(X_valid), np.array(y_valid)

    return (X_train, y_train), (X_valid, y_valid)



def load_histology_shard(verifible_dataset_info, verify_dataset_items,
                         categorical=False, channels_last=False, **kwargs):
    """
    Load the Histology dataset.

    Args:
        data_path (str): path to data directory
        categorical (bool): True = convert the labels to one-hot encoded
         vectors (Default = True)
        channels_last (bool): True = The input images have the channels
         last (Default = True)
        **kwargs: Additional parameters to pass to the function

    Returns:
        list: The input shape
        int: The number of classes
        numpy.ndarray: The training data
        numpy.ndarray: The training labels
        numpy.ndarray: The validation data
        numpy.ndarray: The validation labels
    """
    img_rows, img_cols = 150, 150
    num_classes = 8

    (X_train, y_train), (X_valid, y_valid) = _load_raw_data(verifible_dataset_info, verify_dataset_items, **kwargs)

    if channels_last:
        X_train = X_train.reshape(X_train.shape[0], img_rows, img_cols, 3)
        X_valid = X_valid.reshape(X_valid.shape[0], img_rows, img_cols, 3)
        input_shape = (img_rows, img_cols, 3)
    else:
        X_train = X_train.reshape(X_train.shape[0], 3, img_rows, img_cols)
        X_valid = X_valid.reshape(X_valid.shape[0], 3, img_rows, img_cols)
        input_shape = (3, img_rows, img_cols)

    logger.info(f'Histology > X_train Shape : {X_train.shape}')
    logger.info(f'Histology > y_train Shape : {y_train.shape}')
    logger.info(f'Histology > Train Samples : {X_train.shape[0]}')
    logger.info(f'Histology > Valid Samples : {X_valid.shape[0]}')

    if categorical:
        # convert class vectors to binary class matrices
        y_train = np.eye(num_classes)[y_train]
        y_valid = np.eye(num_classes)[y_valid]

    return input_shape, num_classes, X_train, y_train, X_valid, y_valid



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
import importlib
import logging
import os
import zipfile

from openfl.federated import Plan
from openfl.federated.data.loader import DataLoader

logger = logging.getLogger(__name__)


def initialize_minimal_dataloader(plan: Plan) -> DataLoader:
    """Initialize a minimal dataloader without loading actual data.

    This is used primarily for model initialization when the actual data
    is not needed.

    Args:
        plan (Plan): Plan object linked with the dataloader

    Returns:
        DataLoader: A minimal dataloader instance with no data loaded

    Raises:
        ValueError: If required configuration is missing or dataloader class cannot be found
    """
    # Get the dataloader template from plan
    if "data_loader" not in plan.config or "template" not in plan.config["data_loader"]:
        logger.error("Missing 'data_loader' or 'template' field in plan configuration")
        raise ValueError("Invalid plan configuration: missing data_loader template")

    dataloader_template = plan.config["data_loader"]["template"]

    # Dynamically import the dataloader class
    module_name, class_name = dataloader_template.rsplit(".", 1)
    module = importlib.import_module(module_name)
    if not hasattr(module, class_name):
        logger.error(f"Class {class_name} not found in module {module_name}")
        raise ValueError(
            f"Cannot load data_loader class '{class_name}' from module '{module_name}'"
        )

    dataloader_class = getattr(module, class_name)

    # Initialize dataloader with None as data_path to skip data loading
    if "settings" not in plan.config["data_loader"]:
        logger.error("Missing 'settings' field in data_loader configuration")
        raise ValueError("Invalid plan configuration: missing data_loader settings")

    data_loader_settings = plan.config["data_loader"]["settings"].copy()
    data_loader = dataloader_class(data_path=None, **data_loader_settings)
    logger.info("Initialized minimal dataloader for model creation")
    return data_loader


def initialize_dataloader(
    plan: Plan,
    collaborator_index: int = 0,
) -> DataLoader:
    """Initialize a dataloader with actual data from a collaborator.

    NOTE: cwd must be the workspace directory because we need to
    construct dataloader from actual collaborator data path
    with actual data present.

    Args:
        plan (Plan): Plan object linked with the dataloader
        collaborator_index (int): Which collaborator should be used for initializing dataloader
            among collaborators specified in plan/data.yaml. Defaults to 0.

    Returns:
        DataLoader: A dataloader instance with data loaded

    Raises:
        ValueError: If collaborator_index is out of range
    """
    # Regular dataloader initialization with actual data paths
    collaborator_names = list(plan.cols_data_paths)
    collaborators_count = len(collaborator_names)

    if collaborator_index >= collaborators_count:
        raise ValueError(
            f"Unable to construct full dataloader from collab_index={collaborator_index} "
            f"when the plan has {collaborators_count} as total collaborator count. "
            f"Please check plan/data.yaml file for current collaborator entries."
        )

    collaborator_name = collaborator_names[collaborator_index]
    collaborator_data_path = plan.cols_data_paths[collaborator_name]

    # use seed_data provided by data_loader config if available
    if "seed_data" in plan.config["data_loader"]["settings"] and not os.path.isdir(
        collaborator_data_path
    ):
        os.makedirs(collaborator_data_path)
        sample_data_zip_file = plan.config["data_loader"]["settings"]["seed_data"]
        with zipfile.ZipFile(sample_data_zip_file, "r") as zip_ref:
            zip_ref.extractall(collaborator_data_path)

    data_loader = plan.get_data_loader(collaborator_name)
    return data_loader



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from importlib import util

if util.find_spec("tensorflow") is not None:
    from openfl.utilities.optimizers.keras.fedprox import FedProxOptimizer  # NOQA



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""FedCurv Aggregation function module."""

import numpy as np

from openfl.interface.aggregation_functions.weighted_average import WeightedAverage


class FedCurvWeightedAverage(WeightedAverage):
    """Aggregation function of FedCurv algorithm.

    Applies weighted average aggregation to all tensors
    except Fisher matrices variables (u_t, v_t).
    These variables are summed without weights.

    FedCurv paper: https://arxiv.org/pdf/1910.07796.pdf
    """

    def call(self, local_tensors, tensor_db, tensor_name, fl_round, tags):
        """Apply aggregation."""
        if tensor_name.endswith("_u") or tensor_name.endswith("_v") or tensor_name.endswith("_w"):
            tensors = [local_tensor.tensor for local_tensor in local_tensors]
            agg_result = np.sum(tensors, axis=0)
            return agg_result
        return super().call(local_tensors)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""gRPC interceptors module."""

import collections

import grpc


class _GenericClientInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
    grpc.StreamUnaryClientInterceptor,
    grpc.StreamStreamClientInterceptor,
):
    def __init__(self, interceptor_function):
        self._fn = interceptor_function

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, False
        )
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response

    def intercept_unary_stream(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, True
        )
        response_it = continuation(new_details, next(new_request_iterator))
        return postprocess(response_it) if postprocess else response_it

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, False
        )
        response = continuation(new_details, new_request_iterator)
        return postprocess(response) if postprocess else response

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, True
        )
        response_it = continuation(new_details, new_request_iterator)
        return postprocess(response_it) if postprocess else response_it


def _create_generic_interceptor(intercept_call):
    return _GenericClientInterceptor(intercept_call)


class _ClientCallDetails(
    collections.namedtuple("_ClientCallDetails", ("method", "timeout", "metadata", "credentials")),
    grpc.ClientCallDetails,
):
    pass


def headers_adder(headers):
    """Create interceptor with added headers."""

    def intercept_call(
        client_call_details,
        request_iterator,
        request_streaming,
        response_streaming,
    ):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        for header, value in headers.items():
            metadata.append(
                (
                    header,
                    value,
                )
            )
        client_call_details = _ClientCallDetails(
            client_call_details.method,
            client_call_details.timeout,
            metadata,
            client_call_details.credentials,
        )
        return client_call_details, request_iterator, None

    return _create_generic_interceptor(intercept_call)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Adam optimizer module."""

from typing import Dict, Optional, Tuple

import numpy as np

from openfl.utilities.optimizers.numpy.adam_optimizer import NumPyAdam


class NumPyYogi(NumPyAdam):
    """Yogi optimizer implementation.

    Implements the Yogi optimization algorithm using NumPy.
    Yogi is an algorithm for first-order gradient-based optimization of
    stochastic objective functions, based on adaptive estimates of lower-order
    moments. It is a variant of Adam and it is more robust to large learning
    rates.

    Original paper:
    https://papers.nips.cc/paper/8186-adaptive-methods-for-nonconvex-optimization

    Attributes:
        params (dict, optional): Parameters to be stored for optimization.
        model_interface: Model interface instance to provide parameters.
        learning_rate (float): Tuning parameter that determines the step size
            at each iteration.
        betas (tuple): Coefficients used for computing running averages of
            gradient and its square.
        initial_accumulator_value (float): Initial value for gradients and
            squared gradients.
        epsilon (float): Value for computational stability.
    """

    def __init__(
        self,
        *,
        params: Optional[Dict[str, np.ndarray]] = None,
        model_interface=None,
        learning_rate: float = 0.01,
        betas: Tuple[float, float] = (0.9, 0.999),
        initial_accumulator_value: float = 0.0,
        epsilon: float = 1e-8,
    ) -> None:
        """Initialize the Yogi optimizer.

        Args:
            params (dict, optional): Parameters to be stored for optimization.
                Defaults to None.
            model_interface: Model interface instance to provide parameters.
                Defaults to None.
            learning_rate (float, optional): Tuning parameter that determines
                the step size at each iteration. Defaults to 0.01.
            betas (tuple, optional): Coefficients used for computing running
                averages of gradient and its square. Defaults to (0.9, 0.999).
            initial_accumulator_value (float, optional): Initial value for
                gradients and squared gradients. Defaults to 0.0.
            epsilon (float, optional): Value for computational stability.
                Defaults to 1e-8.
        """
        super().__init__(
            params=params,
            model_interface=model_interface,
            learning_rate=learning_rate,
            betas=betas,
            initial_accumulator_value=initial_accumulator_value,
            epsilon=epsilon,
        )

    def _update_second_moment(self, grad_name: str, grad: np.ndarray) -> None:
        """Override second moment update rule for Yogi optimization updates.

        Args:
            grad_name (str): The name of the gradient.
            grad (np.ndarray): The gradient values.
        """
        sign = np.sign(grad**2 - self.grads_second_moment[grad_name])
        self.grads_second_moment[grad_name] = (
            self.beta_2 * self.grads_second_moment[grad_name] + (1.0 - self.beta_2) * sign * grad**2
        )

    def step(self, gradients: Dict[str, np.ndarray]) -> None:
        """Perform a single step for parameter update.

        Implement Yogi optimizer weights update rule.

        Args:
            gradients (dict): Partial derivatives with respect to optimized
                parameters.
        """
        super().step(gradients)



# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from bisect import bisect

import matplotlib.pyplot as plt
import numpy as np
import torch

from privacy_meter.audit import Audit
from privacy_meter.constants import InferenceGame
from privacy_meter.dataset import Dataset
from privacy_meter.information_source import InformationSource

import time
from privacy_meter.metric import PopulationMetric
from privacy_meter.information_source_signal import ModelLoss, ModelGradientNorm
import privacy_meter.hypothesis_test as prtest


class PM_report:  # NOQA: N801
    """
    This class indicates the information about the auditing and
    save the history of the privacy loss during the training.
    """

    def __init__(
        self,
        fpr_tolerance_list: list,
        is_report_roc: bool,
        signals: list,
        level: str,
        log_dir: str,
        interval: int = 1,
        other_info: dict = None,
    ) -> None:
        """
        Constructor

        Args:
            fpr_tolerance_list: FPR tolerance value(s) to be used by the audit.
            is_report_roc: Indicate whether the report should include the whole ROC
                           of the privacy loss
            signal: A list which indicates the auditing signal, e.g., [loss, gradient norm].
            level: Indicate which model to audit, e.g., local model or global model
            log_dir: Indicate where to save the privacy loss profile during the training.
            interval: Indicate the auditing interval
            other_info: contains other parameters, e.g., on which layer we want to
                        compute the gradient norm
        """
        self.fpr_tolerance = fpr_tolerance_list
        self.is_report_roc = is_report_roc
        self.signals = signals
        self.level = level  # which target model to audit
        self.log_dir = log_dir
        self.interval = interval
        self.history = {}  # history information based on each snapshot of the model
        self.other_info = other_info

        for attr in ["tpr", "fpr", "auc", "roc", "round"]:
            self.history[attr] = []

    def update_history(self, attr, info):
        self.history[attr].append(info)


def PopulationAuditor(target_model, datasets, pm_info):  # NOQA: N802
    """
    Function that returns updated privacy risk report based on the current
    snapshot of the model and FL history and updates the PM history.

    Args:
        target_model (PM model obj): The current snapshot of the model
        datasets (dict): Dataset dictionary, which contains members dataset and i
                         non-members dataset, as well as the dataset for auditing.
                         Each dataset is the PM dataset obj
        pm_info (PM_report obj): Dictionary that contains the history
                                 of the privacy loss report
    Returns:
        Updated privacy report
    """

    begin_time = time.time()

    train_dataset = datasets["train"]
    test_dataset = datasets["test"]
    population_dataset = datasets["audit"]

    # prepare for the dataset
    if torch.is_tensor(train_dataset.data):
        train_ds = {"x": train_dataset.data, "y": train_dataset.targets}
        test_ds = {"x": test_dataset.data, "y": test_dataset.targets}
        population_ds = {"x": population_dataset.data, "y": population_dataset.targets}
    else:
        train_ds = {
            "x": torch.from_numpy(train_dataset.data).float(),
            "y": torch.tensor(train_dataset.targets),
        }
        test_ds = {
            "x": torch.from_numpy(test_dataset.data).float(),
            "y": torch.tensor(test_dataset.targets),
        }
        population_ds = {
            "x": torch.from_numpy(population_dataset.data).float(),
            "y": torch.tensor(population_dataset.targets),
        }

    target_dataset = Dataset(
        data_dict={"train": train_ds, "test": test_ds},
        default_input="x",
        default_output="y",
        default_group="y",
    )

    # create the reference dataset
    pm_population_dataset = Dataset(
        data_dict={"train": population_ds},
        default_input="x",
        default_output="y",
        default_group="y",
    )

    target_info_source = InformationSource(
        models=[target_model], datasets=[target_dataset]
    )

    reference_info_source = InformationSource(
        models=[target_model], datasets=[pm_population_dataset]
    )

    print(f"debug:prepare for the dataset {time.time() - begin_time}")

    start_time = time.time()
    metrics = []
    for signal in pm_info.signals:
        if signal == "loss":
            metrics.append(
                PopulationMetric(
                    target_info_source=target_info_source,
                    reference_info_source=reference_info_source,
                    signals=[ModelLoss()],
                    hypothesis_test_func=prtest.linear_itp_threshold_func,
                )
            )
        elif signal == "logits":
            metrics.append(
                PopulationMetric(
                    target_info_source=target_info_source,
                    reference_info_source=reference_info_source,
                    signals=[ModelLoss()],
                    hypothesis_test_func=prtest.logit_rescale_threshold_func,
                )
            )
        elif signal == "gradient_norm":
            metrics.append(
                PopulationMetric(
                    target_info_source=target_info_source,
                    reference_info_source=reference_info_source,
                    signals=[
                        ModelGradientNorm(
                            pm_info.other_info["is_features"],
                            pm_info.other_info["layer_number"],
                        )
                    ],
                    hypothesis_test_func=prtest.linear_itp_threshold_func,
                )
            )
        else:
            raise ValueError(f"The provided signal {signal} is not supported.")

    print(f"debug:construct the metrics uses {time.time() - start_time}")

    start_time = time.time()
    audit_obj = Audit(
        metrics=metrics,
        inference_game_type=InferenceGame.PRIVACY_LOSS_MODEL,
        target_info_sources=target_info_source,
        reference_info_sources=reference_info_source,
        fpr_tolerances=None,
        save_logs=False,
    )
    audit_obj.prepare()
    print(f"preparing the auditing uses {time.time() - start_time}")

    start_time = time.time()
    audit_results = audit_obj.run()
    print(f"debug:running the auditing uses {time.time() - start_time}")

    start_time = time.time()
    roc_list = []
    auc_list = []
    t_tpr_dict = []
    t_fpr_dict = []
    for idx in range(len(audit_results)):
        # for each results in the signals
        mr = audit_results[idx][0]
        tpr_list = mr.tp / (mr.tp + mr.fn + 0.0000001)
        fpr_list = mr.fp / (mr.tn + mr.fp + 0.0000001)
        roc_list.append({"fpr": fpr_list, "tpr": tpr_list})
        auc_list.append(np.trapz(x=fpr_list, y=tpr_list))

        # focus on the performance uses are interested in
        tolerance_index = []
        for t_fpr in pm_info.fpr_tolerance:
            p = bisect(fpr_list, t_fpr)
            if p < len(fpr_list):
                tolerance_index.append(p)
            else:
                tolerance_index.append(p - 1)  # avoid to select the end point
        t_tpr_list = tpr_list[tolerance_index]
        t_fpr_list = fpr_list[tolerance_index]
        t_tpr_dict.append(t_tpr_list)
        t_fpr_dict.append(t_fpr_list)

    pm_info.update_history("tpr", t_tpr_dict)
    pm_info.update_history("fpr", t_fpr_dict)

    pm_info.update_history("auc", auc_list)
    pm_info.update_history("roc", roc_list)

    print(f"debug:saving the information uses {time.time() - start_time}")
    return pm_info


def plot_tpr_history(history_dict, client, fpr_tolerance_list):
    """
    Plot the history of true positive rate for each fpr tolerance
    in the fpr_tolerance_list  during the training.

    history_list: contains all the signals or reports we aim to show
    client: the auditing client
    name_list: idx for the history list
    fpr_tolerance_list: indicate the fpr tolerance list
    """
    for idx in range(len(fpr_tolerance_list)):
        for key in history_dict:
            for sidx, signal in enumerate(history_dict[key].signals):
                hist = np.array(history_dict[key].history["tpr"])[:, sidx, idx]
                plt.plot(
                    history_dict[key].history["round"], hist, label=f"{key}-{signal}"
                )
        plt.title(f"{client} (FPR @ {fpr_tolerance_list[idx]})")
        plt.legend()
        plt.savefig(
            f"{history_dict[key].log_dir}/{client}_tpr_at_{fpr_tolerance_list[idx]}.png",
            dpi=200,
        )
        plt.clf()


def plot_auc_history(history_dict, client):
    """
    Plot the history of a metric during the training.
    """

    for key in history_dict:
        for sidx, signal in enumerate(history_dict[key].signals):
            plt.plot(
                history_dict[key].history["round"],
                np.array(history_dict[key].history["auc"])[:, sidx],
                label=f"{key}-{signal}",
            )

    plt.title(f"{client} (AUC)")
    plt.legend()
    plt.savefig(f"{history_dict[key].log_dir}/{client}_auc.png", dpi=200)
    plt.clf()


def plot_roc_history(history_dict, client):
    """
    Plot the history of a metric during the training.
    """

    for key in history_dict:
        for sidx, signal in enumerate(history_dict[key].signals):
            tpr = history_dict[key].history["roc"][-1][sidx]["tpr"]
            fpr = history_dict[key].history["roc"][-1][sidx]["fpr"]
            plt.plot(fpr, tpr, label=f"{key}-{signal}")

        round_num = history_dict[key].history["round"][-1]  # the current round number

    plt.grid()
    plt.legend(loc="upper left")
    plt.xlabel("False positive rate (FPR)")
    plt.ylabel("True positive rate (TPR)")
    plt.xlim(0, 1)
    plt.ylim(0, 1)
    plt.plot([0, 1], [0, 1], ls="--", color="gray", label="Random guess")

    plt.title(f"ROC curve - {client}")
    plt.legend()
    plt.savefig(
        f"{history_dict[key].log_dir}/{client}_roc_at_{round_num}.png",
        dpi=200,
    )
    plt.clf()



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""FedProx Keras optimizer module."""

import tensorflow as tf
import tensorflow.keras as keras


@keras.utils.register_keras_serializable()
class FedProxOptimizer(keras.optimizers.Optimizer):
    """FedProx optimizer (Keras3 based API).

    Implements the FedProx algorithm as a Keras optimizer. FedProx is a
    federated learning optimization algorithm designed to handle non-IID data.
    It introduces a proximal term to the federated averaging algorithm to
    reduce the impact of devices with outlying updates.

    Paper: https://arxiv.org/pdf/1812.06127.pdf

    Attributes:
        learning_rate (float): The learning rate for the optimizer.
        mu (float): The proximal term coefficient.
    """

    def __init__(
        self,
        learning_rate=0.01,
        mu=0.0,
        name="FedProxOptimizer",
        **kwargs,
    ):
        super().__init__(
            learning_rate=learning_rate,
            name=name,
            **kwargs,
        )
        self.mu = mu

    def build(self, variables):
        """Initialize optimizer variables.

        Args:
            variables (list): List of model variables to build FedProx variables on.
        """
        if self.built:
            return
        super().build(variables)
        self.vstars = []
        for variable in variables:
            self.vstars.append(
                self.add_variable_from_reference(reference_variable=variable, name="vstar")
            )

    def update_step(self, gradient, variable, learning_rate):
        """Update step given gradient and the associated model variable.
            In the update_step method, variable is updated using the
            gradient and the proximal term (mu). The proximal term helps
            to regularize the update by considering the difference between
            the current value of variable and its initial value (vstar),
            which was stored during the build method.
        Args:
            gradient (tf.Tensor): The gradient tensor for the variable.
            variable (tf.Variable): The model variable to be updated.
            learning_rate (float): The learning rate for the update step.
        """
        lr_t = tf.cast(learning_rate, variable.dtype)
        mu_t = tf.cast(self.mu, variable.dtype)
        gradient_t = tf.cast(gradient, variable.dtype)
        # Get the corresponding vstar for the current variable
        vstar = self.vstars[self._get_variable_index(variable)]

        # Update the variable using the gradient and the proximal term
        self.assign_sub(variable, lr_t * (gradient_t + mu_t * (variable - vstar)))

    def get_config(self):
        """Return the config of the optimizer.
        An optimizer config is a Python dictionary (serializable)
        containing the configuration of an optimizer.
        The same optimizer can be reinstantiated later
        (without any saved state) from this configuration.
        Returns:
            dict: The optimizer configuration.
        """
        config = super().get_config()
        config.update(
            {
                "mu": self.mu,
            }
        )
        return config



# Copyright (C) 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import torch, torchvision, random
import numpy as np
from torch.utils.data import TensorDataset

RANDOM_SEED = 10
STD_DEV = torch.from_numpy(np.array([0.2023, 0.1994, 0.2010]))
MEAN = torch.from_numpy(np.array([0.4914, 0.4822, 0.4465]))
BATCH_SIZE_TRAIN = 32
BATCH_SIZE_TEST = 1000

def seed_random_generators(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)

def trigger_single_image(image):
    """
    Adds a red square with a height/width of 6 pixels into
    the upper left corner of the given image.
    @param image tensor, containing the normalized pixel values of the image.
    The image will be modified in-place.
    @return given image
    """
    color = (torch.Tensor((1, 0, 0)) - MEAN) / STD_DEV
    image[:, 0:6, 0:6] = color.repeat((6, 6, 1)).permute(2, 1, 0)
    return image

def poison_data(samples_to_poison, labels_to_poison, pdr=0.5):
    """
    poisons a given local dataset, consisting of samples and labels, s.t.,
    the given ratio of this image consists of samples for the backdoor behavior
    :param samples_to_poison tensor containing all samples of the local dataset
    :param labels_to_poison tensor containing all labels
    :param pdr poisoned data rate
    :return poisoned local dataset (samples, labels)
    """
    if pdr == 0:
        return samples_to_poison, labels_to_poison

    assert 0 < pdr <= 1.0
    samples_to_poison = samples_to_poison.clone()
    labels_to_poison = labels_to_poison.clone()

    dataset_size = samples_to_poison.shape[0]
    num_samples_to_poison = int(dataset_size * pdr)
    if num_samples_to_poison == 0:
        # corner case for tiny pdrs
        assert pdr > 0  # Already checked above
        assert dataset_size > 1
        num_samples_to_poison += 1

    indices = np.random.choice(dataset_size, size=num_samples_to_poison, replace=False)
    for image_index in indices:
        image = trigger_single_image(samples_to_poison[image_index])
        samples_to_poison[image_index] = image
    labels_to_poison[indices] = 2
    return samples_to_poison, labels_to_poison.long()

transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor(), torchvision.transforms.Normalize(MEAN, STD_DEV), ])
cifar_train = torchvision.datasets.CIFAR10(root="../files", train=True, download=True, transform=transform)
cifar_train = list(cifar_train)
cifar_test = torchvision.datasets.CIFAR10(root="../files", train=False, download=True, transform=transform)
cifar_test = list(cifar_test)


def envoy_attrs(index, malicious, number_of_collaborators, train_dataset_ratio, test_dataset_ratio, cifar_train, cifar_test):
    # split the dataset
    seed_random_generators(RANDOM_SEED)
    X = torch.stack([x[0] for x in cifar_train] + [x[0] for x in cifar_test])
    Y = torch.LongTensor(np.stack(np.array([x[1] for x in cifar_train] + [x[1] for x in cifar_test])))
    shuffled_indices = np.arange(X.shape[0])
    np.random.shuffle(shuffled_indices)

    N_total_samples = len(cifar_test) + len(cifar_train)
    train_dataset_size = int(N_total_samples * train_dataset_ratio)
    test_dataset_size = int(N_total_samples * test_dataset_ratio)
    X = X[shuffled_indices]
    Y = Y[shuffled_indices]

    train_dataset_data = X[:train_dataset_size]
    train_dataset_targets = Y[:train_dataset_size]

    test_dataset_data = X[train_dataset_size:train_dataset_size + test_dataset_size]
    test_dataset_targets = Y[train_dataset_size:train_dataset_size + test_dataset_size]
    
    benign_training_x = train_dataset_data[index::number_of_collaborators]
    benign_training_y = train_dataset_targets[index::number_of_collaborators]

    if malicious:
        local_train_data, local_train_targets = poison_data(benign_training_x,
                                                            benign_training_y)
    else:
        local_train_data, local_train_targets = benign_training_x, benign_training_y

    local_test_data = test_dataset_data[index::number_of_collaborators]
    local_test_targets = test_dataset_targets[index::number_of_collaborators]

    poison_test_data, poison_test_targets = poison_data(local_test_data, local_test_targets,
                                                        pdr=1.0)
    
    return {
        "train_loader": torch.utils.data.DataLoader(
            TensorDataset(local_train_data, local_train_targets),
            batch_size=BATCH_SIZE_TRAIN, shuffle=True
        ),
        "test_loader": torch.utils.data.DataLoader(
            TensorDataset(local_test_data, local_test_targets),
            batch_size=BATCH_SIZE_TEST, shuffle=False
        ),
        "backdoor_test_loader": torch.utils.data.DataLoader(
            TensorDataset(poison_test_data, poison_test_targets),
            batch_size=BATCH_SIZE_TEST, shuffle=False
        ),
    }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/experimental/workflow/protocols/aggregator.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7openfl/experimental/workflow/protocols/aggregator.proto\x12\x1copenfl.experiment.aggregator\x1a\x1bopenfl/protocols/base.proto\"o\n\rMessageHeader\x12\x0e\n\x06sender\x18\x01 \x01(\t\x12\x10\n\x08receiver\x18\x02 \x01(\t\x12\x17\n\x0f\x66\x65\x64\x65ration_uuid\x18\x03 \x01(\t\x12#\n\x1bsingle_col_cert_common_name\x18\x04 \x01(\t\"\xae\x01\n\x12TaskResultsRequest\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\x12\x13\n\x0b\x63ollab_name\x18\x02 \x01(\t\x12\x14\n\x0cround_number\x18\x03 \x01(\x05\x12\x11\n\tnext_step\x18\x04 \x01(\t\x12\x1d\n\x15\x65xecution_environment\x18\x05 \x01(\x0c\"R\n\x13TaskResultsResponse\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\"N\n\x0fGetTasksRequest\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\"\xbd\x01\n\x10GetTasksResponse\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x15\n\rfunction_name\x18\x03 \x01(\t\x12\x1d\n\x15\x65xecution_environment\x18\x04 \x01(\x0c\x12\x12\n\nsleep_time\x18\x05 \x01(\x05\x12\x0c\n\x04quit\x18\x06 \x01(\x08\"\x98\x01\n\x11\x43heckpointRequest\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader\x12\x1d\n\x15\x65xecution_environment\x18\x02 \x01(\x0c\x12\x10\n\x08\x66unction\x18\x03 \x01(\x0c\x12\x15\n\rstream_buffer\x18\x04 \x01(\x0c\"Q\n\x12\x43heckpointResponse\x12;\n\x06header\x18\x01 \x01(\x0b\x32+.openfl.experiment.aggregator.MessageHeader2\x84\x02\n\nAggregator\x12U\n\x0fSendTaskResults\x12\x0b.DataStream\x1a\x31.openfl.experiment.aggregator.TaskResultsResponse\"\x00(\x01\x12J\n\x08GetTasks\x12-.openfl.experiment.aggregator.GetTasksRequest\x1a\x0b.DataStream\"\x00\x30\x01\x12S\n\x0e\x43\x61llCheckpoint\x12\x0b.DataStream\x1a\x30.openfl.experiment.aggregator.CheckpointResponse\"\x00(\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'openfl.experimental.workflow.protocols.aggregator_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_MESSAGEHEADER']._serialized_start=118
  _globals['_MESSAGEHEADER']._serialized_end=229
  _globals['_TASKRESULTSREQUEST']._serialized_start=232
  _globals['_TASKRESULTSREQUEST']._serialized_end=406
  _globals['_TASKRESULTSRESPONSE']._serialized_start=408
  _globals['_TASKRESULTSRESPONSE']._serialized_end=490
  _globals['_GETTASKSREQUEST']._serialized_start=492
  _globals['_GETTASKSREQUEST']._serialized_end=570
  _globals['_GETTASKSRESPONSE']._serialized_start=573
  _globals['_GETTASKSRESPONSE']._serialized_end=762
  _globals['_CHECKPOINTREQUEST']._serialized_start=765
  _globals['_CHECKPOINTREQUEST']._serialized_end=917
  _globals['_CHECKPOINTRESPONSE']._serialized_start=919
  _globals['_CHECKPOINTRESPONSE']._serialized_end=1000
  _globals['_AGGREGATOR']._serialized_start=1003
  _globals['_AGGREGATOR']._serialized_end=1263
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from copy import deepcopy

import torch
import torchvision\



# Download Test datasets

mnist_test = torchvision.datasets.MNIST(
    "./files/",
    train=False,
    download=True,
    transform=torchvision.transforms.Compose(
        [
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.1307,), (0.3081,)),
        ]
    ),
)

# shard the dataset according to collaborator index
portland_col_idx = 0
n_collaborators = 2
batch_size_test = 1000

local_test = deepcopy(mnist_test)

local_test.data = mnist_test.data[portland_col_idx::n_collaborators]
local_test.targets = mnist_test.targets[portland_col_idx::n_collaborators]


portland_attrs = {
   'test_loader': torch.utils.data.DataLoader(local_test,batch_size=batch_size_test, shuffle=True)
}



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Implementation of FedCurv algorithm."""

from copy import deepcopy

import torch
import torch.nn.functional as F


def register_buffer(module: torch.nn.Module, name: str, value: torch.Tensor):
    """Add a buffer to a module.

    Args:
        module (torch.nn.Module): The module to add the buffer to.
        name (str): The name of the buffer. Supports complex module names like
            'model.conv1.bias'.
        value (torch.Tensor): The value of the buffer.
    """
    module_path, _, name = name.rpartition(".")
    mod = module.get_submodule(module_path)
    mod.register_buffer(name, value)


def get_buffer(module, target):
    """Get a buffer from a module.

    Remove after pinning to a version
    where https://github.com/pytorch/pytorch/pull/61429 is included.
    Use module.get_buffer() instead.

    Args:
        module (torch.nn.Module): The module to get the buffer from.
        target (str): The name of the buffer to get.

    Returns:
        torch.Tensor: The buffer.
    """
    module_path, _, buffer_name = target.rpartition(".")

    mod: torch.nn.Module = module.get_submodule(module_path)

    if not hasattr(mod, buffer_name):
        raise AttributeError(f"{mod._get_name()} has no attribute `{buffer_name}`")

    buffer: torch.Tensor = getattr(mod, buffer_name)

    if buffer_name not in mod._buffers:
        raise AttributeError("`" + buffer_name + "` is not a buffer")

    return buffer


class FedCurv:
    """Federated Curvature class.

    This class implements the FedCurv algorithm for federated learning.
    Requires torch>=1.9.0.

    Args:
        model (torch.nn.Module): The base model. Parameters of it are used in
            loss penalty calculation.
        importance (float): The lambda coefficient of the FedCurv algorithm.
    """

    def __init__(self, model: torch.nn.Module, importance: float):
        """Initialize the FedCurv object.

        Args:
            model (torch.nn.Module): The base model. Parameters of it are used
                in loss penalty calculation.
            importance (float): The lambda coefficient of the FedCurv
                algorithm.
        """
        self.importance = importance
        self._params = {}
        self._register_fisher_parameters(model)

    def _register_fisher_parameters(self, model):
        """Register the Fisher parameters of the model.

        Args:
            model (torch.nn.Module): The model to register the Fisher
                parameters for.
        """
        params = list(model.named_parameters())
        for n, p in params:
            u = torch.zeros_like(p, requires_grad=False)
            v = torch.zeros_like(p, requires_grad=False)
            w = torch.zeros_like(p, requires_grad=False)

            # Add buffers to model for aggregation
            register_buffer(model, f"{n}_u", u)
            register_buffer(model, f"{n}_v", v)
            register_buffer(model, f"{n}_w", w)

            # Store buffers locally for subtraction in loss function
            setattr(self, f"{n}_u", u)
            setattr(self, f"{n}_v", v)
            setattr(self, f"{n}_w", w)

    def _update_params(self, model):
        """Update the parameters of the model.

        Args:
            model (torch.nn.Module): The model to update the parameters for.
        """
        self._params = deepcopy({n: p for n, p in model.named_parameters() if p.requires_grad})

    def _diag_fisher(self, model, data_loader, device):
        """Calculate the diagonal of the Fisher information matrix.

        Args:
            model (torch.nn.Module): The model to calculate the Fisher
                information matrix for.
            data_loader (Iterable): The data loader for the training data.
            device (str): The device to perform the calculations on.

        Returns:
            dict: The diagonal of the Fisher information matrix.
        """
        precision_matrices = {}
        for n, p in self._params.items():
            p.data.zero_()
            precision_matrices[n] = p.data.to(device)

        model.eval()
        model.to(device)
        for sample, target in data_loader:
            model.zero_grad()
            sample = sample.to(device)
            target = target.to(device)
            output = model(sample)
            loss = F.nll_loss(F.log_softmax(output, dim=1), target)
            loss.backward()

            for n, p in model.named_parameters():
                if p.requires_grad:
                    precision_matrices[n].data = p.grad.data**2 / len(data_loader)

        return precision_matrices

    def get_penalty(self, model):
        """Calculate the penalty term for the loss function.

        Args:
            model (torch.nn.Module): The model to calculate the penalty for.
                Stores global u_t and v_t values as buffers.

        Returns:
            float: The penalty term.
        """
        penalty = 0
        if not self._params:
            return penalty
        for name, param in model.named_parameters():
            if param.requires_grad:
                u_global, v_global, w_global = (
                    get_buffer(model, target).detach()
                    for target in (f"{name}_u", f"{name}_v", f"{name}_w")
                )
                u_local, v_local, w_local = (
                    getattr(self, name).detach() for name in (f"{name}_u", f"{name}_v", f"{name}_w")
                )
                u = u_global - u_local
                v = v_global - v_local
                w = w_global - w_local
                _penalty = param**2 * u - 2 * param * v + w
                penalty += _penalty.sum()
        penalty = self.importance * penalty
        return penalty.float()

    def on_train_begin(self, model):
        """Pre-train steps.

        Args:
            model (torch.nn.Module): The model for training.
        """
        self._update_params(model)

    def on_train_end(self, model: torch.nn.Module, data_loader, device):
        """Perform post-training steps.

        Args:
            model (torch.nn.Module): The trained model.
            data_loader (Iterable): The data loader for the training data.
            device (str): The device that the model was trained on.
        """
        precision_matrices = self._diag_fisher(model, data_loader, device)
        for n, m in precision_matrices.items():
            u = m.data.to(device)
            v = m.data * model.get_parameter(n)
            v = v.to(device)
            w = m.data * model.get_parameter(n) ** 2
            w = w.to(device)
            register_buffer(model, f"{n}_u", u.clone().detach())
            register_buffer(model, f"{n}_v", v.clone().detach())
            register_buffer(model, f"{n}_w", w.clone().detach())
            setattr(self, f"{n}_u", u.clone().detach())
            setattr(self, f"{n}_v", v.clone().detach())
            setattr(self, f"{n}_w", w.clone().detach())
